/*
 *      ISIS release V1.1, Dec. 1988
 *      Export restrictions apply
 */

/*
 * We use this to test "ports" of ISIS to new system types
 * It exercises the task mechanisms in a mild but adequate way
 * and reveals problems saving/restoring registers.  Can be
 * linked to clib and will run even if ISIS is not up
 */

#include <isis.h>

#define		MAXTASKS	32
#define		MAXSWITCH	5000

condition	cond[MAXTASKS];
int             siged[MAXTASKS];

int		switched = 0;
int		next_task = 0;
int             n_tasks = 3;
int		inc = 1;
int		urgent_flag = 0;

TWAIT(n)
  register n;
  {
	register r;
	if(r = siged[n])
	{
	    print("%2d: Waiting (** pre-signaled **).\n", n);
	    siged[n] = 0;
	    return(r-1);
	}
	print("%2d: Waiting.\n", n);
	return(t_wait(&cond[n]));
  }

TSIG(how, next, n)
  register next, n;
  {
	if(t_waiting(&cond[next]) == 0)
	{
	    print("%2d: %s signal %2d (** wasn't waiting **)\n", n, (how && n)? "urgently": "delay", next);
	    siged[next] = n+1;
	    return;
	}
	print("%2d: %s signal %2d\n", n, (how && n)? "urgently": "delay", next);
        if(how && n)
	    t_sig_urgent(&cond[next], n);
        else
	    t_sig(&cond[next], n);
  }

test_task(n)
  {
	register i0,i1,i2,i3,i4,i5,i6,i7,*p0,*p1,*p2,*p3,*p4,*p5,*p6,*p7;
	int who = -1;
	print("%2d: Task starting up.\n", n);
	i0 = n+0; i1=n+1; i2=n+2; i3=n+3; i4=n+4; i5=n+5; i6=n+6; i7=n+7;
	p0=(int*)(n*0x100+0x00); p1=(int*)(n*0x100+0x10); p2=(int*)(n*0x100+0x20); p3=(int*)(n*0x100+0x30); p4=(int*)(n*0x100+0x40); p5=(int*)(n*0x100+0x50); p6=(int*)(n*0x100+0x60); p7=(int*)(n*0x100+0x70);
	if(n != 0)
	    who = TWAIT(n);
        forever
	{
	     printf("%2d: Running.  [%d/%d/%d/%d/%d/%d/%d/%d;%x/%x/%x/%x/%x/%x/%x/%x].  %2d woke me up\n",
		n, i0,i1,i2,i3,i4,i5,i6,i7,p0,p1,p2,p3,p4,p5,p6,p7, who);
	     if((next_task += inc) >= n_tasks)
		next_task -= n_tasks;
	     TSIG(urgent_flag, next_task, n);
	     if(++switched > MAXSWITCH)
		break;
	     who = TWAIT(n);
	     if(n == 0 && (++inc >= n_tasks))
	        inc = 1;
	}
	print("%2d: Done.\n", n);
  }

main(argc, argv)
  char **argv;
  {
	register n;
	if(argc-- > 1)
	{
	    n_tasks = atoi(*++argv);
	    if(n_tasks <= 0)
		panic("Usage");
	    for(n = 2; n < n_tasks; n++)
	        if(n_tasks%n == 0)
		    panic("Usage");
	}
	if(argc > 1)
	    ++urgent_flag;
	print("Task tester: %d tasks, urgent %d\n", n_tasks, urgent_flag);
	isis_init(1533);
	isis_task(test_task, "task_tester: test task");
	for(n = n_tasks-1; n != 0; n--)
	{
            print("Forking task %d\n", n);
#	if(THREADS)
	    if(n&1)
		THREAD_FORK(test_task, n);
	    else
#	endif
		t_fork(test_task, n);
	}
	print("Forking main task\n");
	isis_mainloop(test_task, 0);
	print("Task tester: finished\n");
  }
